;; Generated package description from do-at-point.el  -*- no-byte-compile: t -*-
(define-package "do-at-point" "0.2.0pre0.20260202.180210" "Generic context-sensitive action dispatcher." '((emacs "26.1")) :commit "a3cc684f47bffd8e820533470f392816978dcbcb" :authors '(("Philip Kaludercic" . "philipk@posteo.net")) :maintainer '("Philip Kaludercic" . "philipk@posteo.net") :keywords '("convenience") :url "https://codeberg.org/pkal/do-at-point.el")
